#!/usr/bin/python

#
#Bif-2D Copyright (C) 2016  Marco Spadini
#
#This program comes with ABSOLUTELY NO WARRANTY. 
#This is free software, and you are welcome to redistribute it under certain 
#conditions.  
#See the GNU General Public License for more details.
#
#
#About the author:
#Marco Spadini                         
#Dipartimento di Matematica e Informatica,
#Via S. Marta 3, 50139 Firenze,       
#ITALY 
#
#web: http:\\www.dma.unifi.it\~spadini
#e-mail: marco.spadini@unifi.it
#
#Description:  
#Given a system of two scalar, parameter-dependent differential equation 
#
#x'=f(t,x,y,l),
#y'=g(t,x,y,l) 
#
#where f and g are T-periodic in t, the program attempts to find the triples (l,x_0,y_0) 
#such that the solution of the 2-dimensional differential equation, corresponding to l 
#and starting at time t=0 from (x_0,y_0), is T-periodic.
#
#The vector function (f,g) ought to be inserted as a FORTRAN 90/95 (or higher) function
#that accepts and returns double precision reals. The function must be placed in the 
#'equations' directory. The search is limited to points (l,x_0,y_0) in a box 
#(Lmin,Lmax)X(Xmin,Xmax)X(Ymin,Ymax), where Lmin, Lmax, Xmin, Xmax, Ymin and Ymax 
#must be specified by the user. Promising initial points are fist looked for in a three
#dimensional grid of nx X ny X nl cells (nx, ny and nl should be provided by the user) all 
#of the same size, and then refined. A point (l,x_0,y_0) is considered good if the solution
#at time T of the Cauchy problem
#
#
#x'=f(t,x,y,l),
#y'=g(t,x,y,l),
#x(0)=x_0,
#y(0)=y_0
#
#is closer, in 1-norm, to (x_0,y_0) than a given tolerance. 
#
#The program outputs its raw data in the directory 'tmp' and then uses this data to
#produce different kinds of graphics according to some user-given parameters. There are 
#two data files 'disegno.dat' contains the refined points (the 4th column is the 2-norm
#of the point) and 'disegno2.dat' contains the unrefined ones. Gnuplot scripts 
#generating the grapics prepared according to the given parameters are placed in the
#same 'tmp' directory.
#
#
#Components.
#The program consists of three software pieces:
#1) The command-line interface (bif-2D.py) written in Python;
#2) The FORTRAN main program and modules that actually makes the computations, 
#   all Fortran sources must stay in the directory 'src'.
#
#Requirements:
#A working gfortran installation is needed for the main program. A working Python (I 
#tested v. 2.7) installation is required for the interface. Also, to get graphics output, 
#a working Gnuplot installation with wxt support is required.
#
#
#Installation: 
#Just copy this directory somewhere you have the rights to read write and 
#execute.
#
#CALLING PARAMETERS:
#positional arguments:
#  RHS                   RightHandSide routine name (extension included)

#optional arguments:
#  -h, --help            show this help message and exit
#  -x XRNGE [XRNGE ...], --xrange XRNGE [XRNGE ...]
#                        range for x variable
#  -y YRNGE [YRNGE ...], --yrange YRNGE [YRNGE ...]
#                        range for y variable
#  -l LRNGE [LRNGE ...], --lrange LRNGE [LRNGE ...]
#                        range for l parameter
#  -nx NX, --nstepx NX   # of grid points on x (default = 100)
#  -ny NY, --nstepy NY   # of grid points on y (default = 100)
#  -nl NL, --nstepl NL   # of grid points on l (default = 100)
#  -T T, --Time T        Time translation (default: 1.0)
#  -t TOLL, --toll TOLL  Tolerance (default: 5.e-3)
#  -at ATOLL, --atoll ATOLL
#                        Integrator's absolute tolerance (default: 1.e-4)
#  -rt RTOLL, --rtoll RTOLL
#                        Integrator's relative tolerance (default: 1.e-3)
#  -rp ROUGHPTS, --roughp ROUGHPTS
#                        max # of rough points (default 10000)
#  -gp GOODPTS, --goodp GOODPTS
#                        max # of good points (default 5000)
#  --combine             combine rough and refined plots (default: false)
#  --norm                graphs the norm of initial points against lambda
#                        (default: false)
#  --P, --projections    Shows projections on coordinate planes (default:
#                        false)
#  --R, --rough          Shows unrefined points (default: false, inactive if
#                        --combine is specified)
#  --U, --unsafe         Less safe, but faster, numerical evaluations (default: false)
#
#
#EXAMPLE:
#At the prompt, issue the command 
#"python bif-2D.py test00.f90 -x 0.0 1.5 -y -1.5 1.5 -l 0.0 1.0 -nx 100 -ny 100 -nl 100 -T 2*pi". 
#
#NOTE:
#Some parameters should be chosen with care, avoiding exagerations. For instance, an excessively
#greedy choice of --atoll or --rtoll forces the integrator to do a lot of useless work. Similarly
#a too small --tol may result in non convergence of the refiner and loss of important information.
#If one suspect this, please compare the rough (unrefined) output. Another delicate point concerns
#the --unsafe choice: the user should avoid it unless it is sure that in the search box the 
#equation makes sense and all its solutions starting in the search box are defined up to T.
#


import argparse
import sys, os 
from math import *

version ="0.99"

descr = """ Bif-2D vers. %s -- Copyright (C) 2016  Marco Spadini.

This program shows the fixed points of the map given by the time 
translation defined by the parameter-dependent ODE whose right-hand-side is given 
by the user in the form of a FORTRAN (90 or higher) routine (its name must be 
passed as first argument), and with parameter varying within given bounds. The 
other parameters must be specified as optional arguments (see below).
""" % (version, )

parser = argparse.ArgumentParser(description=descr)
parser.add_argument("f", metavar="RHS", action="store",
                    help="RightHandSide routine name (extension included)")
parser.add_argument("-x", "--xrange", action="store", type=str, dest="xrnge", 
                    help="range for x variable", nargs='+')
parser.add_argument("-y", "--yrange", action="store", type=str, dest="yrnge", 
                    help="range for y variable", nargs='+')
parser.add_argument("-l", "--lrange", action="store", type=str, dest="lrnge", 
                    help="range for l parameter", nargs='+')
parser.add_argument("-nx", "--nstepx", action="store", type=int, dest="nx",default='100', 
                    help="# of grid points on x (default = 100)")
parser.add_argument("-ny", "--nstepy", action="store", type=int, dest="ny",default='100', 
                    help="# of grid points on y (default = 100)")
parser.add_argument("-nl", "--nstepl", action="store", type=int, dest="nl",default='100', 
                    help="# of grid points on l (default = 100)")
parser.add_argument("-T", "--Time", action="store", type=str, dest="T",default="1.0", 
                    help="Time translation (default: 1.0)")
parser.add_argument("-t", "--toll", action="store", type=str, dest="toll",default="0.005", 
                    help="Tolerance (default: 5.e-3)")
parser.add_argument("-at", "--atoll", action="store", type=str, dest="atoll",default="0.0001", 
                    help="Integrator's absolute tolerance (default: 1.e-4)")
parser.add_argument("-rt", "--rtoll", action="store", type=str, dest="rtoll",default="0.001", 
                    help="Integrator's relative tolerance (default: 1.e-3)")
parser.add_argument("-rp", "--roughp", action="store", type=str, dest="roughpts",default="10000", 
                    help="max # of rough points (default 10000)")
parser.add_argument("-gp", "--goodp", action="store", type=str, dest="goodpts",default="5000", 
                    help="max # of good points (default 5000)")
parser.add_argument("--combine", action="store_true", dest="linee", 
                  help="combine rough and refined plots (default: false)")    
parser.add_argument("--norm", action="store_true", dest="norm", 
                  help="graphs the norm of initial points against lambda (default: false)")  
parser.add_argument("--P","--projections", action="store_true", dest="proj", 
                  help="Shows projections on coordinate planes (default: false)")            
parser.add_argument("--R","--rough", action="store_true", dest="rough", 
                  help="Shows unrefined points (default: false, inactive if --combine is specified)")            
parser.add_argument("--U","--unsafe", action="store_true", dest="unsafe", 
                  help="Less safe, hopefully faster, numerical evaluations (default: false)")               
arg = parser.parse_args()

tolerance = min(eval(arg.toll),max((eval(arg.xrnge[1])-eval(arg.xrnge[0]))/arg.nx,(eval(arg.yrnge[1])-eval(arg.yrnge[0]))/arg.ny))
if (tolerance < eval(arg.toll)) :
  print "Tolerance forced to: ",tolerance

os.system("rm -f ./bin/*")

equaz = './equations/'+arg.f
commonFlags = " -fopenmp -mtune='native' -march='native' -J ./bin"

if (not arg.unsafe):
  intF = "./src/Ode7-safe.f90 "
  specialF = " -O3 -fno-unsafe-math-optimizations -frounding-math -fsignaling-nans "
else:
  intF = "./src/Ode7.f90 "
  specialF = " -Ofast -ffast-math "

toBeLinked1 = "f95 -c "+ intF + specialF + commonFlags + " -o ./bin/intgr.o"
toBeLinked2 = "f95 -c ./src/service.f95 " + specialF + commonFlags + " -o ./bin/serv.o"
toBeCompiled = "f95 ./src/d3D-8.f95 ./bin/intgr.o ./bin/serv.o " + equaz + specialF + commonFlags + " -o ./bin/drawit"
failure =  os.system(toBeLinked1) or os.system(toBeLinked2) or os.system(toBeCompiled)
if failure:
  print 'Compilation Error'
  quit()
else:
  if not os.path.exists('./tmp/'):
    os.system('mkdir ./tmp')
  os.system('rm ./tmp/* 2> /dev/null')
  err = os.system('./bin/drawit %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s' % (eval(arg.xrnge[0]), 
                  eval(arg.xrnge[1]),eval(arg.yrnge[0]),eval(arg.yrnge[1]),
                  eval(arg.lrnge[0]),eval(arg.lrnge[1]),arg.nx,arg.ny,arg.nl,
                  eval(arg.T),tolerance,eval(arg.atoll),eval(arg.rtoll),
                  eval(arg.roughpts),eval(arg.goodpts)))
  if err:
    print 'Execution error'
    quit()
  else:
    comandicomuni = '''set terminal wxt enhanced\n
    set xyplane 0\n 
    set xrange [%s:%s] \n 
    set yrange [%s:%s] \n 
    set zrange [%s:%s] \n
    set xlabel '{/Times-Italic=18 x}'\n
    set ylabel '{/Times-Italic=18 y}'\n
    set zlabel '{/Symbol=18 l}'\n''' % (eval(arg.xrnge[0]), 
                  eval(arg.xrnge[1]),eval(arg.yrnge[0]),eval(arg.yrnge[1]),
                  eval(arg.lrnge[0]),eval(arg.lrnge[1]) )
    f1 = open('./tmp/disegno.gnuplot', 'w')
    f1.write(comandicomuni)
    if arg.linee:
      f1.write("""set title 'Refined and Unrefined points'\n
	splot './tmp/disegno2.dat' w point ps 0.2 pt 7 title 'Unrefined', './tmp/disegno.dat' w point ps 0.4 pt 7 title 'Refined'\n""")
    else:  
      f1.write("""unset title\n
	splot './tmp/disegno.dat' w points ps 0.3 pt 7 notitle\n""")
      if arg.rough:
        f2 = open('./tmp/disegno2.gnuplot', 'w')
        f2.write(comandicomuni)
        f2.write("""set title 'Unrefined'\n
	            splot './tmp/disegno2.dat' w points ps 0.3 pt 7 notitle\n""")
        f2.close()
        os.system("gnuplot -persist ./tmp/disegno2.gnuplot")
    if arg.norm:
      f2 = open('./tmp/disegno3.gnuplot', 'w')
      f2.write("set terminal wxt enhanced\n set xlabel '{/Symbol=18 l}'\n set ylabel '{/Times-Italic=18 |(x,y)|}'\n")
      f2.write("plot './tmp/disegno.dat' using 3:4  w points ps 0.3 pt 7 notitle\n")
      f2.close()
      os.system("gnuplot -persist ./tmp/disegno3.gnuplot")
    if arg.proj:
      f2 = open('./tmp/pr_xy.gnuplot', 'w')
      f2.write("""set terminal wxt enhanced\n
	set xrange [%s:%s] \n 
        set yrange [%s:%s] \n
        set xlabel '{/Times-Italic=18 x}'\n
        set ylabel '{/Times-Italic=18 y}'\n
        set grid
        plot './tmp/disegno.dat' using 1:2  w points ps 0.3 pt 7 notitle""" % (eval(arg.xrnge[0]), 
                  eval(arg.xrnge[1]),eval(arg.yrnge[0]),eval(arg.yrnge[1])) )
      f2.close()
      os.system("gnuplot -persist ./tmp/pr_xy.gnuplot")
      f2 = open('./tmp/pr_lx.gnuplot', 'w')
      f2.write("""set terminal wxt enhanced\n
	set xrange [%s:%s] \n 
        set yrange [%s:%s] \n
        set xlabel '{/Symbol=18 l}'\n
        set ylabel '{/Times-Italic=18 x}'\n
        set grid
        plot './tmp/disegno.dat' using 3:1  w points ps 0.3 pt 7 notitle""" % (eval(arg.lrnge[0]), 
                  eval(arg.lrnge[1]),eval(arg.xrnge[0]),eval(arg.xrnge[1])) )
      f2.close()
      os.system("gnuplot -persist ./tmp/pr_lx.gnuplot")
      f2 = open('./tmp/pr_ly.gnuplot', 'w')
      f2.write("""set terminal wxt enhanced\n
	set xrange [%s:%s] \n 
        set yrange [%s:%s] \n
        set xlabel '{/Symbol=18 l}'\n
        set ylabel '{/Times-Italic=18 y}'\n
        set grid
        plot './tmp/disegno.dat' using 3:2  w points ps 0.3 pt 7 notitle""" % (eval(arg.lrnge[0]), 
                  eval(arg.lrnge[1]),eval(arg.yrnge[0]),eval(arg.yrnge[1])) )
      f2.close()
      os.system("gnuplot -persist ./tmp/pr_ly.gnuplot")
        
    f1.close()
    os.system("gnuplot -persist ./tmp/disegno.gnuplot")
    

